//
//  PQMoveFilter.swift
//  GPUImage_iOS
//
//  Created by ak on 2020/8/27.
//  Copyright © 2020 Sunset Lake Software LLC. All rights reserved.
//  功能：支持视频源的  读取音视频buffer 参考PLMovieReader
/*
 1. 不要使用CMSampleBufferCreateCopy对当前的CMSampleBuffer进行拷贝（会导致性能降低）
 2. CMSampleBuffer释放问题， 一定要置为NULL， 否则野指针会导致AVAssetReader后续帧无法读取
 */

/*
 AVAssetReader可用于读取AVAsset媒体资源的轨道数据，支持解码、格式转换、mix等操作。但注意事项也不少：

 AVAssetReader不可重复调用startReading，当出现fail或complete状态后也不能重复调用；
 AVAssetReader做解码的时候，切换后台/来电会失去GPU权限，造成解码失败，AVAssetReader也变成fail状态。异常打断结束后，需要重启reader，并确定reader重启成功，否则需要retry；
 AVAssetReader启动后调用seek时，并不会很精准seek到目标点，一般会比指定的时间早几帧（AVPlayer的精准seek，也有同样的问题），需要记录seek的目标时间点，如果seek后读取出的buffer携带的 pts比seek的目标时间小，需要抛弃该数据；
 AVAssetReaderOutput不可重复添加，也不可在AssetReader调用startReading后添加；
 AVAssetReaderOutput不可在未添加前调用 copyNextSampleBuffer；
 AVAssetReader释放资源时，需要调用cancelReading来释放 AVAsset资源，否则会出现fetch不到该资源的问题；
 AVAssetReader对文件视频首帧非关键帧的视频会解码失败，这说明AVAssetReader对文件格式要求很严格，不够鲁棒；
 AVAssetReader不支持m3u8文件，回出现读取不到轨道信息的情况，如果需要解析HLS视频，需要使用FFMpeg进行解封装和VideoToolBox解码；
 AVAssetReader内部创建了解码器和缓存列表，但解码器数量是有限制的（同AVPlayerItem）。
 当然AVAssetReader 只做demux，不做解码工作时可以避免上述一些问题，但需要自行使用VideoToolBox进行硬解，pixel format转换也得单独处理。

 有多种方法可以创建CMTime实例，不过最常见的方法是使用CMTimeMake函数，指定一个64位的value参数和32位的timescale参数，比如，创建一个代表5s的CMTime表达式有下面几种不同的方式：
 CMTime t1 =CMTimeMake(5, 1);
 CMTime t2 =CMTimeMake(3000, 600);
 CMTime t3 =CMTimeMake(5000, 1000);

 二，seek 操作

 AVAssetReader并不适合频繁随机读取的操作，如果需要频繁seek可能需要别的方式实现。
 在开始读取之前，可以对读取的范围进行设置，当开始读取后不可以修改，只能顺序向后读。
 有两种方案来调整读取范围：

 output中可以设置supportsRandomAccess，当为true时，可以重置读取范围，但需要调用方调用copyNextSampleBuffer，直到该方法返回NULL。
 或者重新初始化一个AVAssetReader来设置读取时间。
 如果尝试第一种方案，需要使用seek，可以尝试每次设置一个不太长的区间，以保证读取完整个区间不会耗时太多，且时间间隔最好以关键帧划分。

 */

import Foundation
import UIKit

class PQMoveFilter: PQBaseFilter {
    public var runBenchmark = false

    public weak var delegate: MovieInputDelegate?

    var yuvConversionShader: ShaderProgram?
    var asset: AVAsset?
    var videoComposition: AVVideoComposition?
    var playAtActualSpeed: Bool = false

    // Time in the video where it should start.
    var requestedStartTime: CMTime?

    // Last sample time that played.
    public private(set) var currentTime: CMTime = .zero

    public var loop: Bool?

    // Progress block of the video with a paramater value of 0-1.
    // Can be used to check video encoding progress. Not called from main thread.
    public var progress: ((Double) -> Void)?

    // 开始绘制
    var isStartReading: Bool = false

    var audioSettings: [String: Any]?

    var movieFramebuffer: Framebuffer?
    public var framebufferUserInfo: [AnyHashable: Any]?

    @Atomic var assetReader: AVAssetReader?

    var moveSticker: PQEditVisionTrackMaterialsModel?

    var videoSize: CGSize = .zero

    // 最后一帧图像数据 CMSampleBuffer 不会 deep copy 所以使用一个CVImageBuffer变量
    var lastImageBuffer: CVImageBuffer?

    // 旋转角度值
    var mImageOrientation: ImageOrientation = .portrait

    var inputSize: GLSize = GLSize(width: 0, height: 0)

    // 缓存帧属性
    var cacheFrameBufferMaxCount: Int = 16
    // 缓存数量
    @Atomic var cacheframeBuffers: Array = Array<CMSampleBuffer>.init()

    /// Use serial queue to ensure that the picture is smooth
    var seekQueue: DispatchQueue!
    deinit {}

    public init(url: URL) {
        super.init(fragmentShader: PassthroughFragmentShader, numberOfInputs: 1)

        do {
            try loadAsset(url: url, videoComposition: nil)

        } catch {
            NXLog(message: "load asset  with error: \(error)")
        }
    }

    public init(movieSticker: PQEditVisionTrackMaterialsModel) {
        super.init(fragmentShader: PassthroughFragmentShader, numberOfInputs: 1)
        moveSticker = movieSticker
        FilterLog(message: "资源裁剪的 开始时间\(moveSticker!.model_in)  结束时间: \(moveSticker!.out)")
        if moveSticker!.videoIsCrop() {
            requestedStartTime = CMTimeMake(value: Int64(moveSticker!.model_in * 600), timescale: 600)
        }

        do {
            // 测试代码
            //            try  loadAsset(url:URL(fileURLWithPath:"22222.MP4", relativeTo:Bundle.main.resourceURL!), videoComposition: nil)
            // locationPath 有可能直接使用系统相册地址 e.g.视频地址 var/mobile/Media/DCIM/125APPLE/IMG_5189.MOV 就不用拼接沙盒地址了
            var videoFilePath = movieSticker.locationPath
            if !videoFilePath.contains("var/mobile/Media/DCIM/") {
                videoFilePath = documensDirectory + videoFilePath
            }
            FilterLog(message: "视频地址 \(String(describing: videoFilePath))")
            try loadAsset(url: URL(fileURLWithPath: videoFilePath), videoComposition: nil)

        } catch {
            NXLog(message: "load asset  with error: \(error)")
        }

        FilterLog(message: " move FILTER 初始化 开始显示时间：\(movieSticker.timelineIn) 结束显示时间：\(movieSticker.timelineOut)  裁剪开始时间:\(movieSticker.model_in)  裁剪结束时间:\(movieSticker.out)  路径:\(String(describing: movieSticker.locationPath)) 时长 \(CMTimeGetSeconds(asset?.duration ?? .zero))")

        startReading()
 
        if #available(iOS 10.0, *) {
            seekQueue = DispatchQueue(label: "PQ.moveFiler.seeking", qos: .default, attributes: .initiallyInactive, autoreleaseFrequency: .never, target: nil)
        } else {
            seekQueue = DispatchQueue(label: "PQ.moveFiler.seeking", qos: .userInteractive, attributes: [], autoreleaseFrequency: .inherit, target: nil)
        }
        if #available(iOS 10.0, *) {
            seekQueue.activate()
        }
 
    }

    public override func newFramebufferAvailable(_ framebuffer: Framebuffer, fromSourceIndex: UInt) {
        super.newFramebufferAvailable(framebuffer, fromSourceIndex: fromSourceIndex)

        let currTime = CMTimeGetSeconds(CMTime(value: framebuffer.timingStyle.timestamp!.value, timescale: framebuffer.timingStyle.timestamp!.timescale))

        FilterLog(message: "1111111111 \(currTime)")
 
    }
 

    open override func renderFrame() {


        let inputFramebuffer: Framebuffer = inputFramebuffers[0]!
        inputSize = inputFramebuffer.sizeForTargetOrientation(.portrait)

//        let currTime = CMTimeGetSeconds(CMTime(value: inputFramebuffer.timingStyle.timestamp!.value, timescale: inputFramebuffer.timingStyle.timestamp!.timescale))

//        enableSeek = inputFramebuffer.userInfo?["enableSeek"] as? Bool ?? false
        BFLog(message: "enableSeek is \(enableSeek)")
        currentTime = CMTime(value: inputFramebuffer.timingStyle.timestamp!.value, timescale: inputFramebuffer.timingStyle.timestamp!.timescale)
        FilterLog(message: "11定帧！duration is currentSampleTime is \(CMTimeGetSeconds(currentTime))")

        renderFramebuffer = sharedImageProcessingContext.framebufferCache.requestFramebufferWithProperties(orientation: mImageOrientation, size: inputSize, stencil: false)

        let textureProperties = InputTextureProperties(textureCoordinates: inputFramebuffer.orientation.rotationNeededForOrientation(mImageOrientation).textureCoordinates(), texture: inputFramebuffer.texture)

        renderFramebuffer.activateFramebufferForRendering()
        clearFramebufferWithColor(backgroundColor)
        renderQuadWithShader(shader, uniformSettings: uniformSettings,
                             vertexBufferObject: sharedImageProcessingContext.standardImageVBO, inputTextures: [textureProperties])
        releaseIncomingFramebuffers()

        if CMTimeGetSeconds(currentTime) >= moveSticker!.timelineIn, CMTimeGetSeconds(currentTime) <= moveSticker!.timelineOut {
            FilterLog(message: "开始显示 movefilter 了 开始\(String(describing: moveSticker?.timelineIn)) 结束 ：\(String(describing: moveSticker?.timelineOut)) currentTime \(CMTimeGetSeconds(currentTime)) 裁剪开始时间:\(moveSticker?.model_in) ")
            // 取出视频每一帧并渲染
            if enableSeek {
                resetRangeTime(startTime: currentTime)
                enableSeek = false
            }
            readNextVideoFrame()

            if !isStartReading {
                isStartReading = false
                startReading()
            }
        } else {
            isStartReading = false
            FilterLog(message: "开始显示 movefilter 了 结束了\(String(describing: moveSticker?.timelineIn))  currentTime  \(CMTimeGetSeconds(currentTime)) 裁剪开始时间:\(moveSticker?.model_in) ")
            FilterLog(message: "不显示 movefilter 了")
            
            assetReader?.cancelReading()
        }
    }

    // 原视频角度类型
    func moveAssetRotation() -> NXGPUImageRotationMode {
        let Angle: Int = PQPHAssetVideoParaseUtil.videoRotationAngle(assert: asset!)
//        FilterLog(message: "原视频素材Angle is \(Angle)")
        // see https://my.oschina.net/NycoWang/blog/904105
        switch Angle {
        case -90, 270:
            return .rotateLeftTextureCoordinates
        case 90:
            return .rotateRightTextureCoordinates
        case 180:
            return .rotate180TextureCoordinates
        //        case 270:
        //            return .rotate270TextureCoordinates
        default:
            return .noRotationTextureCoordinates
        }
    }

    // MARK: -

    public func loadAsset(url: URL, videoComposition: AVVideoComposition?, playAtActualSpeed: Bool = false, loop: Bool = false, audioSettings: [String: Any]? = nil) throws {
        asset = AVURLAsset(url: url, options: avAssertOptions)

        if asset != nil {
            let fps = asset!.tracks(withMediaType: .video).first?.nominalFrameRate
            FilterLog(message: "move filter asset  fps is \(String(describing: fps))")

            self.videoComposition = videoComposition
            self.playAtActualSpeed = playAtActualSpeed
            self.loop = loop
            yuvConversionShader = crashOnShaderCompileFailure("MovieInput") { try sharedImageProcessingContext.programForVertexShader(defaultVertexShaderForInputs(2), fragmentShader: YUVConversionFullRangeFragmentShader) }
            self.audioSettings = audioSettings
        } else { FilterLog(message: "asset is nil") }
    }

    // MARK: -

    // MARK: Internal processing functions

    func createReader() -> AVAssetReader? {
        do {
            let outputSettings: [String: AnyObject] =
                [kCVPixelBufferPixelFormatTypeKey as String: NSNumber(value: Int32(kCVPixelFormatType_420YpCbCr8BiPlanarFullRange))]

            assetReader = try AVAssetReader(asset: asset!)

            let videoTrack: AVAssetTrack = asset!.tracks(withMediaType: .video).first!

            videoSize = videoTrack.naturalSize
            FilterLog(message: "视频大小为 ： \(videoSize)")

            if videoComposition == nil {
                let readerVideoTrackOutput = AVAssetReaderTrackOutput(track: asset!.tracks(withMediaType: .video).first!, outputSettings: outputSettings)
                readerVideoTrackOutput.alwaysCopiesSampleData = false
                assetReader!.add(readerVideoTrackOutput)
            } else {
                let readerVideoTrackOutput = AVAssetReaderVideoCompositionOutput(videoTracks: asset!.tracks(withMediaType: .video), videoSettings: outputSettings)
                readerVideoTrackOutput.videoComposition = videoComposition
                readerVideoTrackOutput.alwaysCopiesSampleData = false
                assetReader!.add(readerVideoTrackOutput)
            }

            if let requestedStartTime = self.requestedStartTime {
//                if enableSeek {
                    let outTimeSeconds = moveSticker?.out == 0 ? moveSticker!.duration : moveSticker?.out
//                    assetReader!.timeRange = CMTimeRange(start: requestedStartTime, duration: CMTimeMake(value: Int64((outTimeSeconds ?? 0 - CMTimeGetSeconds(requestedStartTime)) * 600.0), timescale: 600))
                
                assetReader!.timeRange = CMTimeRange(start: CMTime.init(value: Int64((moveSticker?.model_in ?? 0) * 600), timescale: 600), duration: CMTimeMake(value: Int64(((moveSticker?.out ?? 0) - (moveSticker?.model_in ?? 0)) * 600.0), timescale: 600))
                
                

//                } else {
//                    assetReader!.timeRange = CMTimeRange(start: requestedStartTime, duration: CMTimeMake(value: Int64(
//                        // 这里判断有问题?
//                        (moveSticker!.timelineIn == 0 && moveSticker!.timelineOut == 0) ? moveSticker!.duration : (moveSticker!.out - moveSticker!.model_in) * 600
//                    ), timescale: 600))
//                }

                FilterLog(message: "set   assetReader!.timeRange is \(assetReader!.timeRange)")
            }
            return assetReader
        } catch {
            debugPrint("ERROR: Unable to create asset reader: \(error)")
        }
        return nil
    }

    open func startReading() {
        FilterLog(message: "开始初始化")
//        if isStartReading {
//            FilterLog(message: "已经初始化过了")
//            return
//        }

        assetReader?.cancelReading()

        isStartReading = true
        guard let assetReader = createReader() else {
            return // A return statement in this frame will end thread execution.
        }

        do {
            try NSObject.catchException {
                guard assetReader.startReading() else {
                    #if DEBUG
                        cShowHUB(superView: nil, msg: "\(String(describing: assetReader.error))")
                    #endif
                    debugPrint("ERROR: Unable to start reading: \(String(describing: assetReader.error))")
                    return
                }
            }
        } catch {
            debugPrint("ERROR: Unable to start reading: \(error)")
            return
        }
    }

    // 设置解码开始时间
    func resetRangeTime(startTime: CMTime = .zero) {
        FilterLog(message: "\(String(describing: moveSticker?.locationPath)) 取帧的时间 \(CMTimeGetSeconds(requestedStartTime ?? .zero))")
        requestedStartTime = startTime
        startReading()
    }

    // 取出第一帧数据
    func readNextVideoFrame() {
        if assetReader == nil {
            FilterLog(message: "assetReader is error 出现严重错误!!!!!!!!!!!!!!")
            return
        }

        var videoTrackOutput: AVAssetReaderOutput?
        for output in assetReader!.outputs {
            if output.mediaType == AVMediaType.video {
                videoTrackOutput = output
            }
        }
   

        let beginDecoderTime: TimeInterval = Date().timeIntervalSince1970

        let sampleBuffer = videoTrackOutput!.copyNextSampleBuffer()
        let endDecoderTime: TimeInterval = Date().timeIntervalSince1970
        FilterLog(message: "\(beginDecoderTime) \(endDecoderTime)解帧时长  \( TimeInterval(endDecoderTime - beginDecoderTime))")
        if sampleBuffer == nil && assetReader?.status == .completed {
            FilterLog(message: "已经播放完成了 \(CMTimeGetSeconds(currentTime))")
            // 如果不是自动循环模式 且 定帧/裁剪模式 播放完一次后自动停止

            if moveSticker?.materialDurationFit?.fitType != adapterMode.loopAuto.rawValue, moveSticker?.materialDurationFit?.fitType != adapterMode.staticFrame.rawValue {
                return
            }
            // 自动循环模式 重头开始循环
            if moveSticker?.materialDurationFit?.fitType == adapterMode.loopAuto.rawValue {
                FilterLog(message: "自动循环模式 重头开始循环 \(CMTimeGetSeconds(currentTime))")
                isStartReading = false
                startReading()

            } else if moveSticker?.materialDurationFit?.fitType == adapterMode.staticFrame.rawValue {
                if lastImageBuffer != nil {
//                    FilterLog(message: "处理显示定帧")
//                    let currTime = CMTimeGetSeconds(currentTime!)
//                    FilterLog(message: "process time is \(currTime)")
                    process(movieFrame: lastImageBuffer!, withSampleTime: currentTime)
                }

                return
            }

            return
        } else {
            FilterLog(message: "copy sampleBuffer is error ??? \(String(describing: assetReader?.status)) \(sampleBuffer)")
        }

        if sampleBuffer != nil {
            if enableSeek {
                BFLog(message: "cacheframeBuffers 添加后 个数\(cacheframeBuffers.count)")
                cacheframeBuffers.append(sampleBuffer!)
            } else {
                // 正常处理每一帧
                processCurrentBuffer(sampleBuffer: sampleBuffer!)
            }
        }
    }

    func processCurrentBuffer(sampleBuffer: CMSampleBuffer) {
        let currentSampleTime = CMSampleBufferGetOutputPresentationTimeStamp(sampleBuffer)
        let duration = asset!.duration // Only used for the progress block so its acuracy is not critical

        sharedImageProcessingContext.runOperationSynchronously {
            self.process(movieFrame: sampleBuffer)

            //            资源裁剪的 开始时间18.51261146496816  结束时间: 24.29780254777071
            // message: 资源裁剪的 开始时间23.9121237307506  结束时间: 27.537509554140126👈
//            FilterLog(message: "素材 \(self.moveSticker?.locationPath) 取出每一帧的时间: \(self.moveSticker!.out == 0 ? duration.seconds : self.moveSticker!.out) currentSampleTime is \(CMTimeGetSeconds(currentSampleTime))")

            FilterLog(message: "seek 时间为: 素材 \(String(describing: self.moveSticker?.locationPath)) 取出每一帧 显示时间: \(CMTimeGetSeconds(self.currentTime)) 帧时间 \(CMTimeGetSeconds(currentSampleTime))")

            // 视频filter 已经播放完一次了， 设置定帧数据 使用精准时间？ INT
            let outTime: Float64 = self.moveSticker!.out == 0 ? duration.seconds : self.moveSticker!.out
            if (CMTimeGetSeconds(currentSampleTime) - outTime) < 0.033 && self.moveSticker?.materialDurationFit?.fitType == adapterMode.staticFrame.rawValue && lastImageBuffer == nil {
                FilterLog(message: "设置了定帧！！！")
                lastImageBuffer = CMSampleBufferGetImageBuffer(sampleBuffer)!

            } else {
                FilterLog(message: "不能设置定帧！！！")
                CMSampleBufferInvalidate(sampleBuffer)
            }
        }
    }

    func process(movieFrame frame: CMSampleBuffer) {
        let movieFrame = (CMSampleBufferGetImageBuffer(frame) == nil) ? lastImageBuffer : CMSampleBufferGetImageBuffer(frame)!
        if movieFrame != nil {
            process(movieFrame: movieFrame!, withSampleTime: currentTime)
        }
    }

    func process(movieFrame: CVPixelBuffer, withSampleTime: CMTime) {
        let bufferHeight = CVPixelBufferGetHeight(movieFrame)
        let bufferWidth = CVPixelBufferGetWidth(movieFrame)
        CVPixelBufferLockBaseAddress(movieFrame, CVPixelBufferLockFlags(rawValue: CVOptionFlags(0)))

        let conversionMatrix = colorConversionMatrix601FullRangeDefault

        var luminanceGLTexture: CVOpenGLESTexture?

        glActiveTexture(GLenum(GL_TEXTURE0))

        let luminanceGLTextureResult = CVOpenGLESTextureCacheCreateTextureFromImage(kCFAllocatorDefault, sharedImageProcessingContext.coreVideoTextureCache, movieFrame, nil, GLenum(GL_TEXTURE_2D), GL_LUMINANCE, GLsizei(bufferWidth), GLsizei(bufferHeight), GLenum(GL_LUMINANCE), GLenum(GL_UNSIGNED_BYTE), 0, &luminanceGLTexture)

        if luminanceGLTextureResult != kCVReturnSuccess || luminanceGLTexture == nil {
            debugPrint("ERROR: Could not create LuminanceGLTexture")
            return
        }

        let luminanceTexture = CVOpenGLESTextureGetName(luminanceGLTexture!)

        glBindTexture(GLenum(GL_TEXTURE_2D), luminanceTexture)
        glTexParameterf(GLenum(GL_TEXTURE_2D), GLenum(GL_TEXTURE_WRAP_S), GLfloat(GL_CLAMP_TO_EDGE))
        glTexParameterf(GLenum(GL_TEXTURE_2D), GLenum(GL_TEXTURE_WRAP_T), GLfloat(GL_CLAMP_TO_EDGE))

        let luminanceFramebuffer: Framebuffer
        do {
            luminanceFramebuffer = try Framebuffer(context: sharedImageProcessingContext, orientation: .portrait, size: GLSize(width: GLint(bufferWidth), height: GLint(bufferHeight)), textureOnly: true, overriddenTexture: luminanceTexture)
        } catch {
            debugPrint("ERROR: Could not create a framebuffer of the size (\(bufferWidth), \(bufferHeight)), error: \(error)")
            return
        }
        luminanceFramebuffer.timingStyle = .videoFrame(timestamp: Timestamp(withSampleTime))

        var chrominanceGLTexture: CVOpenGLESTexture?

        glActiveTexture(GLenum(GL_TEXTURE1))

        let chrominanceGLTextureResult = CVOpenGLESTextureCacheCreateTextureFromImage(kCFAllocatorDefault, sharedImageProcessingContext.coreVideoTextureCache, movieFrame, nil, GLenum(GL_TEXTURE_2D), GL_LUMINANCE_ALPHA, GLsizei(bufferWidth / 2), GLsizei(bufferHeight / 2), GLenum(GL_LUMINANCE_ALPHA), GLenum(GL_UNSIGNED_BYTE), 1, &chrominanceGLTexture)

        if chrominanceGLTextureResult != kCVReturnSuccess || chrominanceGLTexture == nil {
            debugPrint("ERROR: Could not create ChrominanceGLTexture")
            return
        }

        let chrominanceTexture = CVOpenGLESTextureGetName(chrominanceGLTexture!)

        glBindTexture(GLenum(GL_TEXTURE_2D), chrominanceTexture)
        glTexParameterf(GLenum(GL_TEXTURE_2D), GLenum(GL_TEXTURE_WRAP_S), GLfloat(GL_CLAMP_TO_EDGE))
        glTexParameterf(GLenum(GL_TEXTURE_2D), GLenum(GL_TEXTURE_WRAP_T), GLfloat(GL_CLAMP_TO_EDGE))

        let chrominanceFramebuffer: Framebuffer
        do {
            chrominanceFramebuffer = try Framebuffer(context: sharedImageProcessingContext, orientation: .portrait, size: GLSize(width: GLint(bufferWidth), height: GLint(bufferHeight)), textureOnly: true, overriddenTexture: chrominanceTexture)
        } catch {
            debugPrint("ERROR: Could not create a framebuffer of the size (\(bufferWidth), \(bufferHeight)), error: \(error)")
            return
        }
        chrominanceFramebuffer.timingStyle = .videoFrame(timestamp: Timestamp(withSampleTime))

        self.movieFramebuffer?.unlock()
        let movieFramebuffer = sharedImageProcessingContext.framebufferCache.requestFramebufferWithProperties(orientation: .portrait, size: GLSize(width: GLint(bufferWidth), height: GLint(bufferHeight)), textureOnly: true)
        movieFramebuffer.lock()

        convertYUVToRGBAK(shader: yuvConversionShader!, luminanceFramebuffer: luminanceFramebuffer, chrominanceFramebuffer: chrominanceFramebuffer, resultFramebuffer: movieFramebuffer, colorConversionMatrix: conversionMatrix)
        CVPixelBufferUnlockBaseAddress(movieFrame, CVPixelBufferLockFlags(rawValue: CVOptionFlags(0)))

        FilterLog(message: "process time is \(withSampleTime)")
        movieFramebuffer.timingStyle = .videoFrame(timestamp: Timestamp(withSampleTime))

        movieFramebuffer.userInfo = framebufferUserInfo
        self.movieFramebuffer = movieFramebuffer
    }

    public func convertYUVToRGBAK(shader: ShaderProgram, luminanceFramebuffer: Framebuffer, chrominanceFramebuffer: Framebuffer, secondChrominanceFramebuffer: Framebuffer? = nil, resultFramebuffer: Framebuffer, colorConversionMatrix: Matrix3x3) {
        let textureProperties: [InputTextureProperties]
        if let secondChrominanceFramebuffer = secondChrominanceFramebuffer {
            textureProperties = [luminanceFramebuffer.texturePropertiesForTargetOrientation(resultFramebuffer.orientation), chrominanceFramebuffer.texturePropertiesForTargetOrientation(resultFramebuffer.orientation), secondChrominanceFramebuffer.texturePropertiesForTargetOrientation(resultFramebuffer.orientation)]
        } else {
            textureProperties = [luminanceFramebuffer.texturePropertiesForTargetOrientation(resultFramebuffer.orientation), chrominanceFramebuffer.texturePropertiesForTargetOrientation(resultFramebuffer.orientation)]
        }

        var cropTextureProperties: [InputTextureProperties] = Array()
        for texture in textureProperties {
            let textureCoordinates = PQGPUImageTools.getTextureCoordinates(sticker: moveSticker!, textureSize: videoSize, rotationMode: moveAssetRotation(), cannvasSize: inputSize)

            let texturePropertiesimagetwo = InputTextureProperties(textureCoordinates: textureCoordinates, texture: texture.texture)
            cropTextureProperties.append(texturePropertiesimagetwo)
        }

        // 清空背景色
        // clearFramebufferWithColor(Color(red:0, green:0, blue:0, alpha:1.0))
        let uniformSettings = ShaderUniformSettings()
        uniformSettings["colorConversionMatrix"] = colorConversionMatrix

        let verticesPoint: [GLfloat] = PQGPUImageTools.getVerticesPoint(sticker: moveSticker!, textureSize: (moveAssetRotation() == .rotateLeftTextureCoordinates || moveAssetRotation() == .rotateRightTextureCoordinates) ? CGSize(width: videoSize.height, height: videoSize.width) : videoSize, cannvasSize: inputSize)

        renderQuadWithShader(shader,
                             uniformSettings: uniformSettings,
                             vertexBufferObject: PQGPUImageTools.NXGenerateVBO(for: verticesPoint),
                             inputTextures: cropTextureProperties)
        releaseIncomingFramebuffers()

        luminanceFramebuffer.unlock()
        chrominanceFramebuffer.unlock()
        secondChrominanceFramebuffer?.unlock()
    }
}
